/* Copyright(C) 2004-2005 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "senna_in.h"
#include <stdio.h>                /* for debug */
#include <string.h>
#include <stddef.h>
#include "snip.h"

#if !defined MAX
#define MAX(a, b) ((a) > (b) ? (a) : (b))
#endif

#if !defined MIN
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#endif

/* These BM/Turbo-BM algolithm is implemented using examples from */
/* EXACT STRING MATCHING ALGORITHMS by Christian Charras - Thierry Lecroq */
/* e-mails: {Christian.Charras, Thierry.Lecroq}@laposte.net */
/* http://www-igm.univ-mlv.fr/~lecroq/string/node14.html */
/* AND Fussy's Homepage */
/* e-mail: fushimi@sd.starcat.ne.jp */
/* http://www2.starcat.ne.jp/~fussy/algo/algo7-4.htm */

#define MAX_BM_SEARCH_COUNT	512

int
sen_bm_check_euc(const unsigned char *x, const size_t y)
{
  const unsigned char *p;
  for (p = x + y - 1; p >= x && *p >= 0x80; p--);
  return (int)((x + y - p) % 2);
}

int
sen_bm_check_sjis(const unsigned char *x, const size_t y)
{
  const unsigned char *p;
  for (p = x + y - 1; p >= x; p--)
    if ((*p < 0x81) || (*p > 0x9F && *p < 0xE0) || (*p > 0xEF))
      break;
  return (int)((x + y - p) % 2);
}

void
sen_bm_suffixes(const unsigned char *x, size_t m, size_t *suff)
{
  size_t f, g;
  ptrdiff_t i;
  f = 0;
  suff[m - 1] = m;
  g = m - 1;
  for (i = m - 2; i >= 0; --i) {
    if (i > (ptrdiff_t)g && suff[i + m - 1 - f] < i - g)
      suff[i] = suff[i + m - 1 - f];
    else {
      if (i < (ptrdiff_t)g)
        g = i;
      f = i;
      while (g >= 0 && x[g] == x[g + m - 1 - f])
        --g;
      suff[i] = f - g;
    }
  }
}

#ifdef TBM
void
sen_bm_preBmBc(const unsigned char *x, size_t m, size_t *bmBc)
{
  size_t i;
  for (i = 0; i < ASIZE; ++i)
    bmBc[i] = m;
  for (i = 0; i < m - 1; ++i)
    bmBc[(int) x[i]] = m - i - 1;
}
int
sen_bm_preBmGs(const unsigned char *x, size_t m, size_t *bmGs)
{
  size_t i, j, *suff;
  if (!(suff = SEN_MALLOC(sizeof(size_t) * m))) {
    return 1;
  }
  sen_bm_suffixes(x, m, suff);
  for (i = 0; i < m; ++i)
    bmGs[i] = m;
  j = 0;
  for (i = m - 1; i >= -1; --i)
    if (i == -1 || suff[i] == i + 1)
      for (; j < m - 1 - i; ++j)
        if (bmGs[j] == m)
          bmGs[j] = m - 1 - i;
  for (i = 0; i <= m - 2; ++i)
    bmGs[m - 1 - suff[i]] = m - 1 - i;
  SEN_FREE(suff);
  return 0;
}
int
sen_bm_TBM(_snip_cond * cond, sen_nstr * object, size_t n,
           sen_encoding encoding, int total_count, _bm_search_result * results)
{
  const unsigned char *x, *y;
  unsigned int count = 0;
  size_t offset = 0;
  ptrdiff_t i, j, k, l, m, u, v, lastj, shift, turboShift, bcShift;

  x = (unsigned char *) cond->keyword->norm;
  y = (unsigned char *) object->norm;
  m = (ptrdiff_t)cond->keyword_len;
  j = lastj = l = u = 0;
  shift = m;
  while (j <= (ptrdiff_t)(n - m)) {
    i = m - 1;
    while (i >= 0 && x[i] == y[i + j]) {
      --i;
      if (u != 0 && i == m - 1 - shift)
        i -= u;
    }
    if (i < 0) {
      if (MAX_BM_SEARCH_COUNT <= count + total_count + 1){
        break;
      } else if (object->checks[j]) {
        for (k = lastj; k < j; k++) {
          if (object->checks[k] > 0) {
            l = k;
            offset += object->checks[k];
          }
        }
        if (object->checks[j] < 0){
          offset -= object->checks[l];
          lastj = l;
        }
        else{
          lastj = j;
        }
        results[count].start_offset = results[count].end_offset = offset;
        for (k = lastj; k < lastj + (ptrdiff_t)m; k++) {
          if (object->checks[k] > 0) {
            results[count].end_offset += object->checks[k];
          }
        }
        results[count++].cond = cond;
      }
      shift = cond->bmGs[0];
      u = m - shift;
    } else {
      v = m - 1 - i;
      turboShift = u - v;
      bcShift = cond->bmBc[(int) y[i + j]] - m + 1 + i;
      shift = MAX(turboShift, bcShift);
      shift = MAX(shift, cond->bmGs[i]);
      if (shift == cond->bmGs[i])
        u = MIN(m - shift, v);
      else {
        if (turboShift < bcShift)
          shift = MAX(shift, u + 1);
        u = 0;
      }
    }
    j += shift;
  }
  cond->count = count;
  return count;
}
#endif /* TBM */

#ifdef QUICK
void
sen_bm_preQsBc(const unsigned char *x, size_t m, size_t *qsBc)
{
  size_t i;

  for (i = 0; i < ASIZE; ++i)
    qsBc[i] = m + 1;
  for (i = 0; i < m; ++i)
    qsBc[x[i]] = m - i;
}

int
sen_bm_quick(_snip_cond * cond, sen_nstr * object, size_t n,
           sen_encoding encoding, int total_count, _bm_search_result * results)
{
  const unsigned char *x;
  unsigned char *y;
  unsigned int count = 0;
  size_t offset, *qsBc;
  ptrdiff_t j, k, l, lastj;
  const size_t m = cond->keyword_len;

  x = (unsigned char *) cond->keyword->norm;
  y = (unsigned char *) object->norm;
  qsBc = cond->qsBc;

  /* Searching */
  j = lastj = l = offset = 0;

  while (j <= (ptrdiff_t)(n - m)) {
    if (memcmp(x, y + j, m) == 0){
      if (MAX_BM_SEARCH_COUNT <= count + total_count + 1){
        break;
      } else if (object->checks[j]) {
        for (k = lastj; k < j; k++) {
          if (object->checks[k] > 0) {
            l = k;
            offset += object->checks[k];
          }
        }
        if (object->checks[j] < 0){
          offset -= object->checks[l];
          lastj = l;
        }
        else{
          lastj = j;
        }
        results[count].start_offset = results[count].end_offset = offset;
        for (k = lastj; k < lastj + (ptrdiff_t)m; k++) {
          if (object->checks[k] > 0) {
            results[count].end_offset += object->checks[k];
          }
        }
        results[count++].cond = cond;
      }
    }
    j += qsBc[y[j + m]];               /* shift */
  }
  cond->count = count;
  return count;
}
#endif /* QUICK */

int
count_mapped_chars(const char *str, const char *end)
{
  const char *p;
  int dl;

  dl = 0;
  for (p = str; p != end; p++) {
    switch (*p) {
    case '<':
    case '>':
      dl += 4;
      break;
    case '&':
      dl += 5;
      break;
    case '"':
      dl += 6;
      break;
    default:
      dl++;
      break;
    }
  }
  return dl;
}

sen_rc
snip_cond_close(_snip_cond * cond)
{
  if (!cond) {
    return sen_invalid_argument;
  }
  if (cond->keyword) {
    sen_nstr_close(cond->keyword);
  }
#ifdef TBM
  if (cond->bmGs) {
    SEN_FREE(cond->bmGs);
  }
#endif /* TBM */
  SEN_FREE(cond);
  return sen_success;
}

sen_rc
sen_snip_add_cond(sen_snip * snip, const char *keyword,
                  const char *opentag, const char *closetag)
{
  _snip_cond *cond;
  if (!keyword || !*keyword) {
    return sen_invalid_argument;
  }
  if (!(cond = SEN_MALLOC(sizeof(_snip_cond)))) {
    return sen_memory_exhausted;
  }
  memset(cond, 0, sizeof(_snip_cond));
  if (opentag) {
    cond->opentag = opentag;
    cond->opentag_len = strlen(opentag);
  } else {
    cond->opentag = snip->defaultopentag;
    cond->opentag_len = snip->defaultopentag_len;
  }
  if (closetag) {
    cond->closetag = closetag;
    cond->closetag_len = strlen(closetag);
  } else {
    cond->closetag = snip->defaultclosetag;
    cond->closetag_len = snip->defaultclosetag_len;
  }

  if (snip->flags & SEN_SNIP_NORMALIZE) {
    cond->keyword = sen_nstr_open(keyword, snip->encoding, SEN_STR_REMOVEBLANK);        /* normalize dup */
  } else {
    sen_nstr *nstr;
    if (!(nstr = SEN_MALLOC(sizeof(sen_nstr)))) {
      goto exit;
    }
    nstr->norm = SEN_STRDUP(keyword);        /* dup, the label is 'norm' but not normalized  */
    nstr->ctypes = NULL;
    nstr->checks = NULL;
    cond->keyword = nstr;
  }
  cond->keyword_len = strlen(cond->keyword->norm);

  if (!cond->keyword_len) {
    snip_cond_close(cond);
    return sen_invalid_argument;
  }
#ifdef TBM
  if (!(cond->bmGs = SEN_MALLOC(sizeof(size_t) * cond->keyword_len))) {
    goto exit;
  }
  if (sen_bm_preBmGs(cond->keyword->norm, cond->keyword_len, cond->bmGs)) {
    goto exit;
  }
  sen_bm_preBmBc(cond->keyword->norm, cond->keyword_len, cond->bmBc);
#endif /* TBM */

#ifdef QUICK
  sen_bm_preQsBc(cond->keyword->norm, cond->keyword_len, cond->qsBc);
#endif

  cond->next = NULL;

  if (snip->first_cond) {
    snip->last_cond->next = cond;
    snip->last_cond = cond;
  } else {
    snip->first_cond = snip->last_cond = cond;
  }

  snip->cond_len++;
  return sen_success;

exit:
  snip_cond_close(cond);
  return sen_memory_exhausted;
}

int
offset_compare(_bm_search_result * a, _bm_search_result * b)
{
  if (a->start_offset > b->start_offset) {
    return 1;
  } else if (a->start_offset < b->start_offset) {
    return -1;
  } else {
    if (a->end_offset < b->end_offset) {
      return 1;
    } else if (a->end_offset > b->end_offset) {
      return -1;
    } else {
      return 0;
    }
  }
}

int
snip_score_compare(_snip_result * a, _snip_result * b)
{
  if (a->score < b->score) {
    return 1;
  } else if (a->score > b->score) {
    return -1;
  } else {
    return 0;
  }
}

size_t
sen_snip_find_firstbyte(const char *string, sen_encoding encoding, size_t offset,
                        size_t doffset)
{
  switch (encoding) {
  case sen_enc_euc_jp:
    while (!(sen_bm_check_euc((unsigned char *) string, offset)))
      offset += doffset;
    break;
  case sen_enc_sjis:
    if (!(sen_bm_check_sjis((unsigned char *) string, offset)))
      offset += doffset;
    break;
  case sen_enc_utf8:
    while (string[offset] <= (char) 0xc0)
      offset += doffset;
    break;
  case sen_enc_default:
  case sen_enc_none:
    break;
  }
  return offset;
}

sen_snip *
sen_snip_open(sen_encoding encoding, int flags, size_t width,
              unsigned int max_results, const char *defaultopentag, const char *defaultclosetag,
              sen_snip_mapping * mapping)
{
  sen_snip *ret = NULL;
  if (!(ret = SEN_MALLOC(sizeof(sen_snip)))) {
    return NULL;
  }
  ret->encoding = encoding;
  ret->flags = flags;
  ret->width = width;
  ret->max_results = max_results;
  ret->defaultopentag = defaultopentag;
  ret->defaultclosetag = defaultclosetag;
  ret->defaultopentag_len = strlen(defaultopentag);
  ret->defaultclosetag_len = strlen(defaultclosetag);
  ret->first_cond = NULL;
  ret->last_cond = NULL;
  ret->cond_len = 0;
  ret->mapping = mapping;

  if (!((ret->bmres = SEN_MALLOC(sizeof(_bm_search_result) * MAX_BM_SEARCH_COUNT)))) {
    sen_snip_close(ret);
    return NULL;
  }
  if (!(ret->result = SEN_MALLOC(sizeof(_snip_result) * MAX_BM_SEARCH_COUNT))){
    sen_snip_close(ret);
    return NULL;
  }
  ret->nstr = NULL;

  return ret;
}

sen_rc
exec_clean(sen_snip * snip)
{
  if (snip->nstr) {
    sen_nstr_close(snip->nstr);
    snip->nstr = NULL;
  }
  return sen_success;
}

sen_rc
sen_snip_close(sen_snip * snip)
{
  _snip_cond *cond, *pre_cond;
  if (!snip) {
    return sen_invalid_argument;
  }
  cond = snip->first_cond;
  while (cond) {
    pre_cond = cond;
    cond = cond->next;
    snip_cond_close(pre_cond);
  }
  if (snip->bmres) {
    SEN_FREE(snip->bmres);
  }
  if (snip->result) {
    SEN_FREE(snip->result);
  }
  exec_clean(snip);
  SEN_FREE(snip);
  return sen_success;
}

sen_rc
sen_snip_exec(sen_snip * snip, const char *string, unsigned int *nresults, size_t *max_tagged_len)
{
  int map_length;
  _bm_search_result *bmres = NULL;
  _snip_result *sres = NULL;
  sen_rc ret = sen_success;
  _snip_cond *cond;
  unsigned int max_results = snip->max_results, count;
  size_t nstr_blen, string_blen, i, j, last_end_offset = 0, last_last_end_offset;
  ptrdiff_t so;

  exec_clean(snip);
  bmres = snip->bmres;
  sres = snip->result;
  
  *nresults = 0;
  if (!string || !*string) {
    return sen_invalid_argument;
  }
  string_blen = strlen(string);
  if (snip->flags & SEN_SNIP_NORMALIZE) {
    snip->nstr = sen_nstr_open(string, snip->encoding, SEN_STR_WITH_CHECKS | SEN_STR_REMOVEBLANK);
  } else {
    sen_nstr *nstr;
    int_least8_t f;
    unsigned char c;

    if (!(nstr = SEN_MALLOC(sizeof(sen_nstr)))) {
      ret = sen_memory_exhausted;
      goto exit;
    }
    nstr->norm = SEN_STRDUP(string);
    nstr->ctypes = NULL;
    if (!(nstr->checks = (int_least8_t *) SEN_MALLOC(sizeof(int_least8_t) * string_blen))){
      ret = sen_memory_exhausted;
      goto exit;
    }

    f = 0;
    switch (snip->encoding) {
    case sen_enc_euc_jp:
      for (i = 0; i < string_blen; i++) {
        if (!f) {
          c = (unsigned char) string[i];
          f = ((c >= 0xa1 && c <= 0xfe) || c == 0x8e ? 2 : (c == 0x8f ? 3 : 1)
            );
          nstr->checks[i] = f;
        } else {
          nstr->checks[i] = 0;
        }
        f--;
      }
      break;
    case sen_enc_sjis:
      for (i = 0; i < string_blen; i++) {
        if (!f) {
          c = (unsigned char) string[i];
          f = (c >= 0x81 && ((c <= 0x9f) || (c >= 0xe0 && c <= 0xfc)) ? 2 : 1);
          nstr->checks[i] = f;
        } else {
          nstr->checks[i] = 0;
        }
        f--;
      }
      break;
    case sen_enc_utf8:
      for (i = 0; i < string_blen; i++) {
        if (!f) {
          c = (unsigned char) string[i];
          f = (c & 0x80 ? (c & 0x20 ? (c & 0x10 ? 4 : 3)
                           : 2)
               : 1);
          nstr->checks[i] = f;
        } else {
          nstr->checks[i] = 0;
        }
        f--;
      }
      break;
    default:
      for (i = 0; i < string_blen; i++) {
        nstr->checks[i] = 1;
      }
      break;
    }

    snip->nstr = nstr;
  }

  nstr_blen = strlen(snip->nstr->norm);        /* byte length, not nstr->length */

  for (count = 0, cond = snip->first_cond; cond; cond = cond->next) {
#ifdef QUICK
    count +=
      sen_bm_quick(cond, snip->nstr, nstr_blen, snip->encoding, count, bmres + count);
#endif /* QUICK */
#ifdef TBM
    count +=
      sen_bm_TBM(cond, snip->nstr, nstr_blen, snip->encoding, count, bmres + count);
#endif /* TBM */
  }

  /* offset sort */
  if (count >= 2) {
    qsort(bmres, count, sizeof(_bm_search_result),
          (int (*)(const void *, const void *)) offset_compare);
  }
  *max_tagged_len = 0;
  for (i = 0; i < count; i++) {
    if (bmres[i].start_offset < last_end_offset) {
      sres[i].score = -1;
    } else {
      (*nresults)++;
      sres[i].score = 0;
      sres[i].tagged_len = 0;
      map_length = 0;
      last_last_end_offset = last_end_offset;
      for (j = i; j < count; j++) {
        /* for filter <tag> in case abcd"efghij<klmn>kl"mn</klmn>op */
        if (bmres[j].end_offset > bmres[i].start_offset + snip->width) {
          continue;
        }
        /* for filter <tag2> in case <tag1>a<tag2>b</tag1>c</tag2> */
        /*     filter <tag2> in case <tag1><tag2>a</tag2>b</tag1> */
        if(bmres[j].start_offset >= last_end_offset){
          sres[i].score += (float)1.0f / bmres[j].cond->count;
          last_end_offset = bmres[j].end_offset;
          sres[i].tagged_len += bmres[j].cond->opentag_len + bmres[j].cond->closetag_len;
        }
        else {
          bmres[j].cond = NULL;
        }
      }
      so = 
        MIN((ptrdiff_t)(bmres[i].start_offset + last_end_offset - snip->width) / 2,
            (ptrdiff_t)(string_blen - snip->width));

      if (so < 0 || (size_t)so < last_last_end_offset){
        sres[i].start_offset = last_last_end_offset;
      }
      else{
        sres[i].start_offset = (size_t)so;
      }

      sres[i].start_offset =
        sen_snip_find_firstbyte(string, snip->encoding, sres[i].start_offset, 1);

      sres[i].end_offset = sres[i].start_offset + snip->width;
      sres[i].end_offset =
        sen_snip_find_firstbyte(string, snip->encoding, sres[i].end_offset, -1);
      last_end_offset = sres[i].end_offset;

      sres[i].tagged_len +=
        count_mapped_chars(&string[sres[i].start_offset],
                           &string[sres[i].end_offset]) + 1;
      sres[i].first_cond_id = (unsigned int)i;
      *max_tagged_len = MAX(*max_tagged_len, sres[i].tagged_len);
    }
  }
  *nresults = MIN(*nresults, max_results);
  if (count >= 2) {
  /* score sort */
  qsort(sres, count, sizeof(_snip_result),
        (int (*)(const void *, const void *)) snip_score_compare);
  }

  snip->count = count;
  snip->string = string;

  return ret;
exit:
  exec_clean(snip);
  return ret;
}

sen_rc
sen_snip_get_result(sen_snip * snip, unsigned int index, char *result)
{
  char *p;
  size_t i, j, l, f;

  if (snip->count <= index || !snip->result || !snip->bmres || !snip->nstr) {
    return sen_invalid_argument;
  }
  j = l = f = snip->result[index].first_cond_id;

  for (p = result, i = snip->result[index].start_offset;
       i < snip->result[index].end_offset; i++) {
    while (j < snip->count && snip->bmres[j].start_offset == i) {
      if (snip->bmres[j].end_offset <= snip->result[index].end_offset) {
        memcpy(p, snip->bmres[j].cond->opentag, snip->bmres[j].cond->opentag_len);
        p += snip->bmres[j].cond->opentag_len;
        l = j;
      }
      do{
        j++;
      } while (j < snip->count && !snip->bmres[j].cond);
    }

    if (snip->mapping == (sen_snip_mapping *) - 1) {
      switch (snip->string[i]) {
      case '<':
        *p++ = '&';
        *p++ = 'l';
        *p++ = 't';
        *p++ = ';';
        break;
      case '>':
        *p++ = '&';
        *p++ = 'g';
        *p++ = 't';
        *p++ = ';';
        break;
      case '&':
        *p++ = '&';
        *p++ = 'a';
        *p++ = 'm';
        *p++ = 'p';
        *p++ = ';';
        break;
      case '"':
        *p++ = '&';
        *p++ = 'q';
        *p++ = 'u';
        *p++ = 'o';
        *p++ = 't';
        *p++ = ';';
        break;
      default:
        *p++ = snip->string[i];
        break;
      }
    } else {
      *p++ = snip->string[i];
    }

    if (snip->bmres[l].end_offset == i + 1) {
      memcpy(p, snip->bmres[l].cond->closetag, snip->bmres[l].cond->closetag_len);
      p += snip->bmres[l].cond->closetag_len;
    }
  }
  *p = '\0';

  return sen_success;
}
